/* Copyright (c) 2024 Skyward Experimental Rocketry
 * Author: Federico Lolli
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "Radio.h"

#include <config/RadioConfig.h>
#include <drivers/interrupt/external_interrupts.h>
#include <radio/SX1278/SX1278Frontends.h>

#include <memory>

// === INTERRUPTS ===
#ifdef SX1278_IRQ_DIO0
void __attribute__((used)) SX1278_IRQ_DIO0()
{
    Arpist::Radio::getInstance().handleDioIRQ();
}
#endif

#ifdef SX1278_IRQ_DIO1
void __attribute__((used)) SX1278_IRQ_DIO1()
{
    Arpist::Radio::getInstance().handleDioIRQ();
}
#endif

#ifdef SX1278_IRQ_DIO3
void __attribute__((used)) SX1278_IRQ_DIO3()
{
    Arpist::Radio::getInstance().handleDioIRQ();
}
#endif

namespace Arpist
{

void Radio::handleDioIRQ()
{
    if (started)
    {
        sx1278->handleDioIRQ();
    }
}

bool Radio::sendMsg(const mavlink_message_t &msg)
{
    if (!started)
    {
        return false;
    }

    return mav_driver->enqueueMsg(msg);
}

bool Radio::send(uint8_t *pkt, size_t len) { return sx1278->send(pkt, len); }

ssize_t Radio::receive(uint8_t *pkt, size_t max_len)
{
    return sx1278->receive(pkt, max_len);
}

bool Radio::start(std::function<void(const mavlink_message_t &)> onReceive)
{
    // initialize frontend
    std::unique_ptr<Boardcore::SX1278::ISX1278Frontend> frontend =
        std::make_unique<Boardcore::RA01Frontend>();

    std::unique_ptr<Boardcore::SX1278Fsk> sx1278 =
        std::make_unique<Boardcore::SX1278Fsk>(
            sx1278_bus, radio_cs::getPin(), radio_dio0::getPin(),
            radio_dio1::getPin(), radio_dio3::getPin(),
            Boardcore::SPI::ClockDivider::DIV_64, std::move(frontend));

    // First check if the device is even connected
    if (!sx1278->checkVersion())
        return false;

    // Configure the radio
    if (sx1278->configure(RADIO_CONFIG) != Boardcore::SX1278Fsk::Error::NONE)
        return false;

    // Initialize
    this->sx1278 = std::move(sx1278);

    auto mav_handler =
        [onReceive](RadioMavDriver *channel, const mavlink_message_t &msg)
    { onReceive(msg); };

    mav_driver = std::make_unique<RadioMavDriver>(
        this, mav_handler, Arpist::MAV_SLEEP_AFTER_SEND,
        Arpist::MAV_OUT_BUFFER_MAX_AGE);

    if (!mav_driver->start())
    {
        return false;
    }

    started = true;
    return true;
}

}  // namespace Arpist
