mod cli;
mod fr_api;
mod packet;
mod tm;
mod utils;

use tokio::task::JoinError;

use std::{
    fs::File, io::{self, Read}, path::{Path, PathBuf}, time::{Duration, Instant}
};

use clap::Parser;
use log::{debug, error, info};
use miette::{Context, IntoDiagnostic, Result};
use packet::TimedMessage;
use serialport::SerialPort;
use skyward_mavlink::{
    lyra::MavMessage,
    mavlink::{write_v1_msg, MavHeader},
};
use tm::{read_high_rate_telemetry, read_low_rate_telemetry};

use crate::{
    cli::Cli,
    packet::PacketSequence,
    utils::{get_first_stm32_serial_port, print_stats},
};

const ACK: u8 = 0x06;

#[tokio::main]
async fn main() -> Result<()> {
    let args: Cli = Cli::parse();

    let lvl = if args.verbose {
        log::LevelFilter::Debug
    } else {
        log::LevelFilter::Info
    };

    // initialize logger
    simple_logger::SimpleLogger::new()
        .with_level(lvl)
        .with_utc_timestamps()
        .env()
        .init()
        .into_diagnostic()?;

    if !args.no_sync {
        info!("SYNC mode ENABLED");
    } else {
        info!("SYNC mode DISABLED");
    }

    let port_name = if let Some(name) = args.port {
        name
    } else if let Some(name) = get_first_stm32_serial_port()? {
        debug!("found STM32 device on {}", name);
        name
    } else {
        error!("FATAL: no STM32 serial port found, your device may not be connected properly");
        std::process::exit(1);
    };

    // Open the first serialport available.

    debug!("connecting to serial port {}...", port_name);
    let mut port = serialport::new(port_name, args.baud_rate)
        .timeout(Duration::from_millis(100))
        .open()
        .into_diagnostic()
        .wrap_err("Failed to open serial port")?;
    let mut write_port = port.try_clone().into_diagnostic()?;

    let mut fr = false;

    let mut fr_nr = "".to_string();

    if let Some(nr) = args.fr_api{
        fr_nr = nr;
    }

    if fr_nr.len() > 1 {
        fr = true;
    }

    let pack_csv = get_packets(args.low_rate_file, args.high_rate_file);

    loop{

        let mut packets: Vec<TimedMessage> = Vec::new();

        // In case is specified to use the Flight Radar API for tracking
        if fr {
                let spawned = tokio::spawn(fr_api::api_request(fr_nr.clone()));
                std::thread::sleep(Duration::from_secs(100));
                let (res,):  (Result<Result<Vec<TimedMessage>, ()>, JoinError>,) = tokio::join!(spawned);
                if let Ok(ret) = res {
                    if let Ok(ret_packets) = ret {
                        packets = ret_packets;
                    }
                }
        }
        else {
            if let Ok(ref pck) = pack_csv {packets = pck.to_vec()};
        }
        
        // convert to packet sequence data structure
        let mut packets = PacketSequence::from(packets);
        

        // read from buffer and print to stdout
        let mut buffer = [0u8; 1000];
        let mut start: Option<Instant> = None;
        let mut times = Vec::new();

        // send packet if there is one
        while let Some(packet) = packets.wait_next() {
            if let Some(t) = start {
                times.push(t.elapsed().as_millis());
            }
            start = Some(Instant::now());
            write_port.write_packet(packet.to_owned())?;
            debug!(
                "sent packet ({}/{})",
                packets.sent_packets(),
                packets.total_packets()
            );

            // if sync mode is enabled then wait for ACK
            if !args.no_sync {
                wait_for_ack(&mut port, &mut buffer);
            }
        }

        info!("sent {} packets", packets.sent_packets());
        print_stats(times);

        if !fr {break;}
    }
    
    Ok(())
}

fn get_packets(low_rate_file:Option<PathBuf>, high_rate_file:Option<PathBuf>) -> Result<Vec<TimedMessage>> {
    if let Some(low_rate_file_path) = low_rate_file {
        if let Some(high_rate_file_path) = high_rate_file {
            info!(
                "reading low rate telemetry from file {}",
                low_rate_file_path.display()
            );
            info!(
                "reading high rate telemetry from file {}",
                high_rate_file_path.display()
            );
            let lrtf = File::open(&low_rate_file_path).into_diagnostic()?;
            let hrtf = File::open(&high_rate_file_path).into_diagnostic()?;
            let mut packets = read_low_rate_telemetry(lrtf).into_diagnostic()?;
            packets.extend(read_high_rate_telemetry(hrtf).into_diagnostic()?);
            packets.sort();
            debug!("collected {} packets", packets.len());
            Ok(packets)
        } else {
            error!("No high rate telemetry path specified!");
            std::process::exit(1);
        }
    } else {
        error!("No low rate telemetry path specified!");
        std::process::exit(1);
    }
}

/// Wait for ACK signal from the device
fn wait_for_ack(port: &mut Box<dyn SerialPort>, buffer: &mut [u8]) {
    // wait ACK
    let start = Instant::now();
    loop {
        match port.read(buffer) {
            Ok(t) => {
                // eprint!("{}", String::from_utf8_lossy(&buffer[..t]));
                // if didn't get the signal then print the buffer and exit with error
                if buffer[..t].contains(&ACK) {
                    break;
                } else {
                    std::thread::sleep(Duration::from_millis(10));
                }
            }
            Err(ref e) if e.kind() == io::ErrorKind::TimedOut => {
                if start.elapsed() > Duration::from_secs(5) {
                    error!("FATAL: didn't get the signal");
                    std::process::exit(1);
                }
            }
            Err(e) => {
                error!("FATAL: {:?}", e);
                std::process::exit(1);
            }
        }
    }
    // eprintln!();
}

trait MavLinkPort {
    fn write_packet(&mut self, packet: MavMessage) -> Result<()>;
}

impl<T: AsRef<dyn SerialPort> + std::io::Write> MavLinkPort for T {
    fn write_packet(&mut self, packet: MavMessage) -> Result<()> {
        let header = MavHeader::default();
        write_v1_msg(self, header, &packet)
            .into_diagnostic()
            .wrap_err("Failed to write mavlink message to serial port")?;
        Ok(())
    }
}
