use std::time::{Duration, Instant};

use skyward_mavlink::lyra::MavMessage;

#[derive(Debug, Clone, PartialEq)]
pub struct TimedMessage {
    timestamp: u64,
    msg: MavMessage,
}

impl PartialOrd for TimedMessage {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl Eq for TimedMessage {}

impl Ord for TimedMessage {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        self.timestamp.cmp(&other.timestamp)
    }
}

impl TimedMessage {
    pub fn new(timestamp: u64, msg: MavMessage) -> Self {
        Self { timestamp, msg }
    }
}

pub struct PacketSequence {
    packets: Vec<TimedMessage>,
    index: usize,
    delta_from_start: u64,
    start_instant: Option<Instant>,
}

impl From<Vec<TimedMessage>> for PacketSequence {
    fn from(value: Vec<TimedMessage>) -> Self {
        let delta_from_start = value[0].timestamp;
        Self {
            packets: value,
            index: 0,
            delta_from_start,
            start_instant: None,
        }
    }
}

impl PacketSequence {
    pub fn sent_packets(&self) -> usize {
        self.index
    }

    pub fn total_packets(&self) -> usize {
        self.packets.len()
    }

    pub fn wait_next(&mut self) -> Option<&MavMessage> {
        // return None if we reached the end of the sequence
        if self.index == self.packets.len() {
            return None;
        }

        // get time delta from the start
        let next_timestamp = self.packets[self.index].timestamp;
        let time_delta = next_timestamp - self.delta_from_start;

        // get the instant the first packet was sent or set it if it's the first packet
        if let Some(start_inst) = self.start_instant {
            // get the instant to wait until
            let till = start_inst + Duration::from_micros(time_delta);

            // sleep for the difference between the last timestamp and the current one
            std::thread::sleep(till - Instant::now());
        } else {
            // set the start instant
            self.start_instant = Some(Instant::now());
        }
        self.index += 1;
        self.packets.get(self.index - 1).map(|x| &x.msg)
    }
}
